@@ -1,4 +1,4 @@
-# SMB Perl library, Copyright (C) 2014 Mikhael Goikhman, migo@cpan.org
+# SMB Perl library, Copyright (C) 2014-2018 Mikhael Goikhman, migo@cpan.org
 #
 # This program is free software: you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -29,6 +29,7 @@ use SMB::v2::Command::Close;
 use SMB::v2::Command::QueryDirectory;
 use SMB::v2::Command::Read;
 use SMB::v2::Command::Write;
+use SMB::v2::Command::SetInfo;
 use SMB::Tree;
 
 sub new ($$%) {
@@ -36,9 +37,6 @@ sub new ($$%) {
 	my $share_uri = shift;
 	my %options = @_;
 
-	my $verbose = delete $options{verbose} || 0;
-	$options{quiet} = 1 unless $verbose;
-
 	my $self = $class->SUPER::new(
 		%options,
 		server_id => 0,  # running index
@@ -96,7 +94,7 @@ sub get_curr_connection ($) {
 
 	my $connections = $self->connections;
 	unless (%$connections) {
-		$self->err("Called get_curr_connection when no connections established");
+		$self->err("Called get_curr_connection without established connections");
 		return;
 	}
 
@@ -146,7 +144,7 @@ sub process_request ($$$%) {
 
 	my $response = $self->wait_for_response($connection);
 
-	warn "SMB Error on $command_name response: " . ($response ? sprintf "%x", $response->status : "internal") . "\n"
+	warn "SMB Error on $command_name response: " . ($response ? $response->status_name : "internal") . "\n"
 		if !$no_warn && (!$response || $response->is_error);
 
 	return $response;
@@ -271,12 +269,207 @@ sub _basename ($;$) {
 	return $path =~ /.*\Q$delim\E(.*)/ ? $1 : $path;
 }
 
+sub dnload_file ($$$$) {
+	my $self = shift;
+	my $connection = shift;
+	my $filename = shift // return $self->err("No remote file name to download");
+	my $dst_filename = shift // return $self->err("No local file name to save");
+
+	my $response = $self->process_request($connection, 'Create',
+		file_name => $filename,
+	);
+	return unless $response && $response->is_success;
+
+	my $file = $response->openfile->file;
+	my $fid = $response->fid;
+	my $remaining = $file->end_of_file;
+	my $time = $file->mtime;
+	my $content = '';
+	my $offset = 0;
+	while ($remaining) {
+		my $length = $remaining >= 65536 ? 65536 : $remaining;
+		$remaining -= $length;
+		$response = $self->process_request($connection, 'Read',
+			fid => $fid,
+			offset => $offset,
+			length => $length,
+			remaining_bytes => $remaining,
+		);
+		return unless $response && $response->is_success;
+		my $read = $response->length;
+		return $self->err("Unexpected $read bytes read instead of $length at offset $offset")
+			if $read != $length;
+		$content .= $response->buffer;
+		$offset += $length;
+	}
+	$self->process_request($connection, 'Close',
+		fid => $fid,
+	);
+
+	open DST, '>', $dst_filename
+		or return $self->err("Can't open $dst_filename for write: $!");
+	print DST $content
+		or return $self->err("Can't write content to $dst_filename: $!");
+	close DST
+		or return $self->err("Can't close $dst_filename after write: $!");
+
+	# consider to set $time on file
+	return 1;
+}
+
+sub upload_file ($$$$) {
+	my $self = shift;
+	my $connection = shift;
+	my $filename = shift // return $self->err("No local file name to load");
+	my $dst_filename = shift // return $self->err("No remote file name to upload");
+
+	local $/ = undef;
+	open SRC, '<', $filename
+		or return $self->err("Can't open $filename for read: $!");
+	my $content = <SRC>
+		// return $self->err("Can't read content from $filename: $!");
+	close SRC
+		or return $self->err("Can't close $filename after read: $!");
+
+	my $response = $self->process_request($connection, 'Create',
+		file_name => $dst_filename,
+		options => SMB::v2::Command::Create::OPTIONS_NON_DIRECTORY_FILE,
+		access_mask => 0x12019f,
+		disposition => SMB::File::DISPOSITION_OVERWRITE_IF,
+	);
+	return unless $response && $response->is_success;
+	my $fid = $response->fid;
+	my $remaining = length($content);
+	my $offset = 0;
+	while ($remaining) {
+		my $length = $remaining >= 65536 ? 65536 : $remaining;
+		$remaining -= $length;
+		$response = $self->process_request($connection, 'Write',
+			fid => $fid,
+			offset => $offset,
+			remaining_bytes => $remaining,
+			buffer => substr($content, $offset, $length),
+		);
+		return unless $response && $response->is_success;
+		my $written = $response->length;
+		return $self->err("Unexpected $written bytes written instead of $length at offset $offset")
+			if $written != $length;
+		$offset += $length;
+	}
+	$self->process_request($connection, 'Close',
+		fid => $fid,
+	);
+
+	return 1;
+}
+
+sub remove_file ($$$$) {
+	my $self = shift;
+	my $connection = shift;
+	my $file = shift // return $self->err("No file to remove");
+	my $recursive = shift;
+
+	my $remove_using_setinfo = $ENV{SMB_CLIENT_REMOVE_FILE_USING_SETINFO};
+
+	my $options = $file->is_directory
+		? SMB::v2::Command::Create::OPTIONS_DIRECTORY_FILE
+		: SMB::v2::Command::Create::OPTIONS_NON_DIRECTORY_FILE;
+	$options |= SMB::v2::Command::Create::OPTIONS_DELETE_ON_CLOSE
+		unless $remove_using_setinfo;
+	my $response = $self->process_request($connection, 'Create',
+		file_name => $file->name,
+		options => $options,
+		access_mask => 0x10081,
+	);
+	return unless $response && $response->is_success;
+	my $fid = $response->fid;
+
+	if ($remove_using_setinfo) {
+		$response = $self->process_request($connection, 'SetInfo',
+			fid => $fid,
+			type => SMB::v2::Command::SetInfo::TYPE_FILE,
+			level => SMB::v2::Command::SetInfo::FILE_LEVEL_DISPOSITION,
+			buffer => chr(SMB::v2::Command::SetInfo::FILE_DISPOSITION_DELETE_ON_CLOSE),
+		);
+		return unless $response && $response->is_success;
+	}
+
+	if ($recursive && $file->is_directory) {
+		my @files = ();
+		while (1) {
+			$response = $self->process_request($connection, 'QueryDirectory',
+				file_pattern => "*",
+				fid => $fid,
+			);
+			last if $response && $response->status == SMB::STATUS_NO_MORE_FILES;
+			return $self->err("Failed to get file list in " . $file->name)
+				unless $response && $response->is_success;
+			push @files, @{$response->files};
+		}
+		my $dirname = $file->name;
+		for my $file (@files) {
+			# TODO: consider to have full file name already on parse-response
+			$file->name("$dirname\\" . $file->name) if $dirname;
+			next if $file->name =~ m/(^|\\)\.\.?$/;
+			return $self->err("Failed to remove inner ". $file->name)
+				unless $self->remove_file($connection, $file, 1);
+		}
+	}
+
+	$self->process_request($connection, 'Close',
+		fid => $fid,
+	);
+	return unless $response && $response->is_success;
+
+	return 1;
+}
+
+sub rename_file ($$$$;$) {
+	my $self = shift;
+	my $connection = shift;
+	my $filename1 = shift // return $self->err("No old filename to rename");
+	my $filename2 = shift // return $self->err("No new filename to rename");
+	my $force = shift || 0;
+
+	my $response = $self->process_request($connection, 'Create',
+		file_name => $filename1,
+		options => 0,
+		access_mask => 0x10081,
+	);
+	return unless $response && $response->is_success;
+	my $fid = $response->fid;
+
+	my $rename_struct = SMB::Packer->new
+		->uint8($force ? 1 : 0)
+		->zero(7)    # reserved
+		->zero(8)    # root dir handle
+		->uint16(length($filename2) * 2)
+		->uint16(0)  # reserved
+		->str($filename2);
+
+	$response = $self->process_request($connection, 'SetInfo',
+		fid => $fid,
+		type => SMB::v2::Command::SetInfo::TYPE_FILE,
+		level => SMB::v2::Command::SetInfo::FILE_LEVEL_RENAME,
+		buffer => $rename_struct->data,
+	);
+	return unless $response && $response->is_success;
+
+	$self->process_request($connection, 'Close',
+		fid => $fid,
+	);
+	return unless $response && $response->is_success;
+
+	return 1;
+}
+
 sub perform_tree_command ($$$@) {
 	my $self = shift;
 	my $tree = shift;
 	my $command = shift;
 
 	my $connection = $self->find_connection_by_tree($tree) || return;
+	my %options = @_ && ref($_[0]) eq 'HASH' ? %{shift()} : ();
 
 	if ($command eq 'chdir') {
 		my $dir = shift // '';
@@ -302,6 +495,7 @@ sub perform_tree_command ($$$@) {
 			fid => $fid,
 		);
 
+		return unless $files;
 		return wantarray ? @$files : $files;
 	} elsif ($command eq 'dnload') {
 		my $filename = shift // '';
@@ -309,112 +503,49 @@ sub perform_tree_command ($$$@) {
 		$filename = _normalize_path($filename, $tree->cwd, 1);
 		my $dst_filename = _normalize_path(shift || _basename($filename, 1), '.');
 
-		my $response = $self->process_request($connection, 'Create',
-			file_name => $filename,
-		);
-		return unless $response && $response->is_success;
-		my $file = $response->openfile->file;
-		my $fid = $response->fid;
-		my $remaining = $file->end_of_file;
-		my $time = $file->mtime;
-		my $content = '';
-		my $offset = 0;
-		while ($remaining) {
-			my $length = $remaining >= 65536 ? 65536 : $remaining;
-			$remaining -= $length;
-			$response = $self->process_request($connection, 'Read',
-				fid => $fid,
-				offset => $offset,
-				length => $length,
-				remaining_bytes => $remaining,
-			);
-			return unless $response && $response->is_success;
-			my $read = $response->length;
-			return $self->err("Unexpected $read bytes read instead of $length at offset $offset")
-				if $read != $length;
-			$content .= $response->buffer;
-			$offset += $length;
-		}
-		$self->process_request($connection, 'Close',
-			fid => $fid,
-		);
-
-		open DST, '>', $dst_filename
-			or return $self->err("Can't open $dst_filename for write: $!");
-		print DST $content
-			or return $self->err("Can't write content to $dst_filename: $!");
-		close DST
-			or return $self->err("Can't close $dst_filename after write: $!");
-
-		# consider to set $time on file
-		return 1;
+		return $self->dnload_file($connection, $filename, $dst_filename);
 	} elsif ($command eq 'upload') {
 		my $filename = shift // '';
 		return $self->err("No filename") if $filename eq '';
 		$filename = _normalize_path($filename, '.');
 		my $dst_filename = _normalize_path(shift || _basename($filename), $tree->cwd, 1);
 
-		local $/ = undef;
-		open SRC, '<', $filename
-			or return $self->err("Can't open $filename for read: $!");
-		my $content = <SRC>
-			// return $self->err("Can't read content from $filename: $!");
-		close SRC
-			or return $self->err("Can't close $filename after read: $!");
-
-		my $response = $self->process_request($connection, 'Create',
-			file_name => $dst_filename,
-			options => SMB::v2::Command::Create::OPTIONS_NON_DIRECTORY_FILE,
-			access_mask => 0x12019f,
-			disposition => SMB::File::DISPOSITION_OVERWRITE_IF,
-		);
-		return unless $response && $response->is_success;
-		my $fid = $response->fid;
-		my $remaining = length($content);
-		my $offset = 0;
-		while ($remaining) {
-			my $length = $remaining >= 65536 ? 65536 : $remaining;
-			$remaining -= $length;
-			$response = $self->process_request($connection, 'Write',
-				fid => $fid,
-				offset => $offset,
-				remaining_bytes => $remaining,
-				buffer => substr($content, $offset, $length),
-			);
-			return unless $response && $response->is_success;
-			my $written = $response->length;
-			return $self->err("Unexpected $written bytes written instead of $length at offset $offset")
-				if $written != $length;
-			$offset += $length;
-		}
-		$self->process_request($connection, 'Close',
-			fid => $fid,
-		);
-
-		return 1;
+		return $self->upload_file($connection, $filename, $dst_filename);
 	} elsif ($command eq 'remove') {
 		my $filename = shift // '';
 		return $self->err("No filename") if $filename eq '';
 		$filename = _normalize_path($filename, $tree->cwd, 1);
 
-		my $is_dir = shift || 0;
-		my $options = ($is_dir
-			? SMB::v2::Command::Create::OPTIONS_DIRECTORY_FILE
-			: SMB::v2::Command::Create::OPTIONS_NON_DIRECTORY_FILE
-		) | SMB::v2::Command::Create::OPTIONS_DELETE_ON_CLOSE;
-		my $response = $self->process_request($connection, 'Create',
-			file_name => $filename,
-			options => $options,
-			access_mask => 0x10081,
-		);
-		return unless $response && $response->is_success;
-		my $fid = $response->fid;
-		$self->process_request($connection, 'Close',
-			fid => $fid,
-		);
-		return unless $response && $response->is_success;
+		my $recursive = $options{recursive};
+		my $is_dir = shift // $recursive;
+		my $file = SMB::File->new(name => $filename, is_directory => $is_dir);
 
-		return 1;
+		return $self->remove_file($connection, $file, $recursive);
+	} elsif ($command eq 'rename') {
+		my $filename1 = shift // '';
+		return $self->err("No filename1") if $filename1 eq '';
+		$filename1 = _normalize_path($filename1, $tree->cwd, 1);
+		my $filename2 = shift // '';
+		return $self->err("No filename2") if $filename2 eq '';
+		$filename2 = _normalize_path($filename2, $tree->cwd, 1);
+		my $force = $options{force};
+
+		return $self->rename_file($connection, $filename1, $filename2, $force);
+	} elsif ($command eq 'copy') {
+		my $filename1 = shift // '';
+		return $self->err("No filename1") if $filename1 eq '';
+		$filename1 = _normalize_path($filename1, $tree->cwd, 1);
+		my $filename2 = shift // '';
+		return $self->err("No filename2") if $filename2 eq '';
+		$filename2 = _normalize_path($filename2, $tree->cwd, 1);
+
+		my $tmp_filename = "/var/tmp/copy-$$";
+		my $success =
+			$self->dnload_file($connection, $filename1, $tmp_filename) &&
+			$self->upload_file($connection, $tmp_filename, $filename2);
+		unlink $tmp_filename;
+
+		return $success;
 	}
 
 	return;
@@ -443,7 +574,7 @@ sub wait_for_response ($$) {
 	}
 
 	unless ($response->is_response_to($request)) {
-		$self->err("Unexpected: " . $response->dump);
+		$self->err("Unexpected: " . $response->to_string);
 		return;
 	}
 
